# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass
from pydantic.typing import Literal

from kolena._api.v1.generic import TestRun as API
from kolena._utils.datatypes import get_args
from kolena._utils.datatypes import get_origin
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import GroundTruth
from kolena.workflow import Inference
from kolena.workflow import TestCase
from kolena.workflow import TestSample
from kolena.workflow import TestSuite
from kolena.workflow._datatypes import DataObject
from kolena.workflow._datatypes import DataType
from kolena.workflow._datatypes import TypedDataObject
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_scalar_data_object_type


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestSample(DataObject, metaclass=ABCMeta):
    """
    Test-sample-level metrics produced by an :class:`Evaluator`.

    This class should be subclassed with the relevant fields for a given workflow.

    Examples here may include the number of true positive detections on an image, the mean IOU of inferred polygon(s)
    with ground truth polygon(s), etc.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_sample_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestCase(DataObject, metaclass=ABCMeta):
    """
    Test-case-level metrics produced by an :class:`Evaluator`.

    This class should be subclassed with the relevant fields for a given workflow.

    Test-case-level metrics are aggregate metrics like Precision, Recall, and F1 score. Any and all aggregate metrics
    that fit a workflow should be defined here.

    ``MetricsTestCase`` supports nesting metrics objects, for e.g. reporting class-level metrics within a test case that
    contains multiple classes. Example usage:

    .. code-block:: python

        @dataclass(frozen=True)
        class PerClassMetrics(MetricsTestCase):
            Class: str
            Precision: float
            Recall: float
            F1: float
            AP: float

        @dataclass(frozen=True)
        class TestCaseMetrics(MetricsTestCase):
            macro_Precision: float
            macro_Recall: float
            macro_F1: float
            mAP: float
            PerClass: List[PerClassMetrics]
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_case_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestSuite(DataObject, metaclass=ABCMeta):
    """
    Test-suite-level metrics produced by an :class:`Evaluator`.

    This class should be subclassed with the relevant fields for a given workflow.

    Test-suite-level metrics typically measure performance across test cases, e.g. penalizing variance across different
    subsets of a benchmark.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_suite_type(cls)


NumberSeries = Sequence[Union[float, int]]
NullableNumberSeries = Sequence[Union[float, int, None]]


class _PlotType(DataType):
    CURVE = "CURVE"
    CONFUSION_MATRIX = "CONFUSION_MATRIX"
    HISTOGRAM = "HISTOGRAM"
    BAR = "BAR"

    @staticmethod
    def _data_category() -> str:
        return "PLOT"


@dataclass(frozen=True, config=ValidatorConfig)
class AxisConfig(DataObject):
    """Configuration for the format of a given axis on a Plot"""

    #: Type of axis to display. Supported options are `linear` and `log`.
    type: Literal["linear", "log"]


@dataclass(frozen=True, config=ValidatorConfig)
class Plot(TypedDataObject[_PlotType], metaclass=ABCMeta):
    """A data visualization shown when exploring model results in the web platform."""


@dataclass(frozen=True, config=ValidatorConfig)
class Curve(DataObject):
    """A single series on a :class:`CurvePlot`."""

    x: NumberSeries
    y: NumberSeries

    #: Optionally specify an additional label (in addition to the associated test case) to apply to this curve, for use
    #: when e.g. there are multiple curves generated per test case.
    label: Optional[str] = None

    def __post_init_post_parse__(self) -> None:
        if len(self.x) != len(self.y):
            raise ValueError(
                f"Series 'x' (length: {len(self.x)}) and 'y' (length: {len(self.y)}) have different lengths",
            )


@dataclass(frozen=True, config=ValidatorConfig)
class CurvePlot(Plot):
    """
    A plot visualizing one or more curves per test case.

    Examples include Receiver Operating Characteristic (ROC) curves, Precision versus Recall (PR) curves,
    Detection-Error Tradeoff (DET) curves, etc.
    """

    title: str
    x_label: str
    y_label: str

    #: A test case may generate zero or more curves on a given plot. However, under most circumstances, a single curve
    #: per test case is desirable.
    curves: List[Curve]

    #: Custom format options to allow for control over the display of the plot axes.
    x_config: Optional[AxisConfig] = None
    y_config: Optional[AxisConfig] = None

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.CURVE


@dataclass(frozen=True, config=ValidatorConfig)
class Histogram(Plot):
    """
    A plot visualizing distribution of one or more continuous values, e.g. distribution of an error metric across all
    samples within a test case.

    For visualization of discrete values, see :class:`BarPlot`.
    """

    title: str
    x_label: str
    y_label: str

    #: A Histogram requires intervals to bucket the data. For ``n`` buckets, ``n+1`` consecutive bounds must be
    #: specified in increasing order.
    buckets: NumberSeries

    #: For ``n`` buckets, there are ``n`` frequencies corresponding to the height of each bucket. The frequency at index
    #: ``i`` corresponds to the bucket with bounds (``i``, ``i+1``) in ``buckets``.
    #:
    #: To specify multiple distributions for a given test case, multiple frequency series can be provided, corresponding
    #: e.g. to the distribution for a given class within a test case, with name specified in ``labels``.
    frequency: Union[NumberSeries, Sequence[NumberSeries]]

    #: Specify a list of labels corresponding to the different ``frequency`` series when multiple series are provided.
    #: Can be omitted when a single ``frequency`` series is provided.
    labels: Optional[List[str]] = None

    #: Custom format options to allow for control over the display of the plot axes.
    x_config: Optional[AxisConfig] = None
    y_config: Optional[AxisConfig] = None

    def __post_init_post_parse__(self) -> None:
        n_buckets = len(self.buckets)
        if n_buckets < 2:
            raise ValueError(f"Minimum 2 entries required in 'buckets' series (length: {n_buckets})")
        buckets_arr = np.array(self.buckets)
        if not np.all(buckets_arr[1:] > buckets_arr[:-1]):  # validate strictly increasing
            raise ValueError("Invalid 'buckets' series: series must be strictly increasing")

        frequency_arr = np.array(self.frequency)
        actual_frequency_shape = np.shape(frequency_arr)
        if len(actual_frequency_shape) > 1 and self.labels is None:
            raise ValueError("'labels' are required when multiple 'frequency' series are provided")

        n_labels = len(self.labels or [])
        expected_frequency_shape = (n_labels, n_buckets - 1) if n_labels > 0 else (n_buckets - 1,)
        if actual_frequency_shape != expected_frequency_shape:
            raise ValueError(f"Invalid 'frequency' shape {actual_frequency_shape}: expected {expected_frequency_shape}")

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.HISTOGRAM


@dataclass(frozen=True, config=ValidatorConfig)
class BarPlot(Plot):
    """A plot visualizing a set of bars per test case."""

    title: str

    #: Axis label for the axis along which the bars are laid out (``labels``).
    x_label: str

    #: Axis label for the axis corresponding to bar height (``values``).
    y_label: str

    #: Labels for each bar with corresponding height specified in ``values``.
    labels: Sequence[Union[str, int, float]]

    #: Values for each bar with corresponding label specified in ``labels``.
    values: NullableNumberSeries

    #: Custom format options to allow for control over the display of the numerical plot axis (``values``).
    config: Optional[AxisConfig] = None

    def __post_init_post_parse__(self) -> None:
        n_labels, n_values = len(self.labels), len(self.values)
        if n_labels == 0 or n_values == 0 or n_labels != n_values:
            raise ValueError(
                f"Series 'labels' (length: {n_labels}) and 'values' (length: {n_values}) "
                "must be equal length and non-empty",
            )

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.BAR


@dataclass(frozen=True, config=ValidatorConfig)
class ConfusionMatrix(Plot):
    """
    A confusion matrix. Example:

    .. code-block:: python

        ConfusionMatrix(
            title="Cat and Dog Confusion",
            labels=["Cat", "Dog"],
            matrix=[[90, 10], [5, 95]],
        )

    Yields a confusion matrix of the form:

    .. code-block:: none

                    Predicted

                    Cat   Dog
                   +----+----+
               Cat | 90 | 10 |
        Actual     +----+----+
               Dog |  5 | 95 |
                   +----+----+
    """

    title: str
    labels: List[str]
    matrix: Sequence[NullableNumberSeries]
    x_label: str = "Predicted"
    y_label: str = "Actual"

    def __post_init_post_parse__(self) -> None:
        n_labels = len(self.labels)
        if n_labels < 2:
            raise ValueError(f"At least two labels required: got {n_labels}")
        if len(self.matrix) != n_labels:
            raise ValueError(f"Invalid number of matrix rows: got {len(self.matrix)}, expected {n_labels}")
        for i, row in enumerate(self.matrix):
            if len(row) != n_labels:
                raise ValueError(f"Invalid number of matrix columns in row {i}: {len(row)}, expected {n_labels}")

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.CONFUSION_MATRIX


@dataclass(frozen=True, config=ValidatorConfig)
class EvaluatorConfiguration(DataObject, metaclass=ABCMeta):
    """
    Configuration for an :class:`Evaluator`.

    Example evaluator configurations may specify:

    - Fixed confidence thresholds at which detections are discarded.
    - Different algorithms/strategies used to compute confidence thresholds
        (e.g. "accuracy optimal" for a classification-type workflow).
    """

    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError


class Evaluator(metaclass=ABCMeta):
    """
    An :class:`kolena.workflow.Evaluator` transforms inferences into metrics.

    Metrics are computed at the individual test sample level (:class:`kolena.workflow.MetricsTestSample`), in aggregate
    at the test case level (:class:`kolena.workflow.MetricsTestCase`), and across populations at the test suite level
    (:class:`kolena.workflow.MetricsTestSuite`).

    Test-case-level plots (:class:`kolena.workflow.Plot`) may also be computed.
    """

    configurations: List[EvaluatorConfiguration]

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, configurations: Optional[List[EvaluatorConfiguration]] = None):
        if configurations is not None and len(configurations) == 0:
            raise ValueError("empty configuration list provided, at least one configuration required or 'None'")
        self.configurations = configurations or []
        if len({configuration.display_name() for configuration in self.configurations}) < len(self.configurations):
            raise ValueError("all configurations must have distinct display names")

    def display_name(self) -> str:
        return type(self).__name__

    @abstractmethod
    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> List[Tuple[TestSample, MetricsTestSample]]:
        """
        Compute metrics for every test sample in a test case.

        Must be implemented.

        :param test_case: the test case to which the provided test samples and ground truths belong.
        :param inferences: the test samples, ground truths, and inferences for all entries in a test case.
        :param configuration: the evaluator configuration to use. Empty for implementations that are not configured.
        :return: test-sample-level metrics for each provided test sample.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[MetricsTestSample],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> MetricsTestCase:
        """
        Compute aggregate metrics across a test case.

        Must be implemented.

        :param test_case: the test case in question.
        :param inferences: the test samples, ground truths, and inferences for all entries in a test case.
        :param metrics: the test-sample-level metrics computed by :meth:`Evaluator.compute_test_sample_metrics`.
            Provided in the same order as ``inferences``.
        :param configuration: the evaluator configuration to use. Empty for implementations that are not configured.
        :return: test-case-level metrics for the provided test case.
        """
        raise NotImplementedError

    @validate_arguments(config=ValidatorConfig)
    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[MetricsTestSample],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[List[Plot]]:
        """
        Optionally compute any number of plots to visualize the results for a test case.

        :param test_case: the test case in question
        :param inferences: the test samples, ground truths, and inferences for all entries in a test case.
        :param metrics: the test-sample-level metrics computed by :meth:`Evaluator.compute_test_sample_metrics`.
            Provided in the same order as ``inferences``.
        :param configuration: the evaluator configuration to use. Empty for implementations that are not configured.
        :return: zero or more plots for this test case at this configuration.
        """
        return None  # not required

    @validate_arguments(config=ValidatorConfig)
    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, MetricsTestCase]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[MetricsTestSuite]:
        """
        Optionally compute test-suite-level metrics.

        :param test_suite: the test suite in question
        :param metrics: the test-case-level metrics computed by :meth:`Evaluator.compute_test_case_metrics`
        :param configuration: the evaluator configuration to use. Empty for implementations that are not configured.
        :return: the test-suite-level metrics for this test suite
        """
        return None  # not required


def _validate_metrics_test_sample_type(metrics_test_sample_type: Type[MetricsTestSample]) -> None:
    # TODO: support special structure for ImagePair test sample types?
    validate_data_object_type(metrics_test_sample_type)


def _validate_metrics_test_case_type(metrics_test_case_type: Type[DataObject]) -> None:
    validate_scalar_data_object_type(metrics_test_case_type, supported_list_types=[MetricsTestCase])

    # validate that there is only one level of nesting
    for field_name, field_type in get_data_object_field_types(metrics_test_case_type).items():
        origin = get_origin(field_type)
        if origin is not list:  # only need to check lists, as MetricsTestCase is only allowed in lists
            continue
        # expand e.g. List[Union[MetricsA, MetricsB]] into [MetricsA, MetricsB]
        list_arg_types = [t for arg_type in get_args(field_type) for t in get_args(arg_type) or [arg_type]]
        for arg_type in list_arg_types:
            if arg_type is None:
                raise ValueError(f"Unsupported optional metrics object in field '{field_name}'")
            try:
                validate_scalar_data_object_type(arg_type)
            except ValueError:
                raise ValueError(f"Unsupported doubly-nested metrics object in field '{field_name}'")


def _validate_metrics_test_suite_type(metrics_test_suite_type: Type[DataObject]) -> None:
    validate_scalar_data_object_type(metrics_test_suite_type)


def _maybe_evaluator_configuration_to_api(
    configuration: Optional[EvaluatorConfiguration],
) -> Optional[API.EvaluatorConfiguration]:
    if configuration is None:
        return None
    return API.EvaluatorConfiguration(display_name=configuration.display_name(), configuration=configuration._to_dict())


def _maybe_display_name(configuration: Optional[EvaluatorConfiguration]) -> Optional[str]:
    if configuration is None:
        return None
    return configuration.display_name()


@validate_arguments(config=ValidatorConfig)
def _configuration_description(configuration: Optional[EvaluatorConfiguration]) -> str:
    display_name = _maybe_display_name(configuration)
    return f"(configuration: {display_name})" if display_name is not None else ""



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from typing import Dict
from typing import List
from typing import Tuple

from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow._datatypes import DataType
from kolena.workflow._datatypes import TypedDataObject


class _AnnotationType(DataType):
    BOUNDING_BOX = "BOUNDING_BOX"
    POLYGON = "POLYGON"
    POLYLINE = "POLYLINE"
    KEYPOINTS = "KEYPOINTS"
    BOUNDING_BOX_3D = "BOUNDING_BOX_3D"
    SEGMENTATION_MASK = "SEGMENTATION_MASK"
    BITMAP_MASK = "BITMAP_MASK"
    CLASSIFICATON_LABEL = "LABEL"

    @staticmethod
    def _data_category() -> str:
        return "ANNOTATION"


@dataclass(frozen=True, config=ValidatorConfig)
class Annotation(TypedDataObject[_AnnotationType], metaclass=ABCMeta):
    """
    Where applicable, annotations are visualized in the web platform.

    For example, when viewing images, any annotations present in a :class:`kolena.workflow.TestSample`,
    :class:`kolena.workflow.GroundTruth`, :class:`kolena.workflow.Inference`, or
    :class:`kolena.workflow.MetricsTestSample` are rendered on top of the image.
    """


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox(Annotation):
    """Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices."""

    top_left: Tuple[float, float]
    bottom_right: Tuple[float, float]

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a string
    label.
    """

    label: str


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a float
    score.
    """

    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices, a string
    label, and a float score.
    """

    label: str
    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class Polygon(Annotation):
    """Arbitrary polygon specified by three or more pixel coordinates."""

    points: List[Tuple[float, float]]

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYGON

    def __post_init__(self) -> None:
        if len(self.points) < 3:
            raise ValueError(f"{type(self).__name__} must have at least three points ({len(self.points)} provided)")


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledPolygon(Polygon):
    """Arbitrary polygon specified by three or more pixel coordinates and a string label."""

    label: str


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates and a float score.
    """

    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates with a string label and a float score.
    """

    label: str
    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class Keypoints(Annotation):
    """Array of any number of keypoints specified in pixel coordinates."""

    points: List[Tuple[float, float]]

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.KEYPOINTS


@dataclass(frozen=True, config=ValidatorConfig)
class Polyline(Annotation):
    """Polyline with any number of vertices specified in pixel coordinates."""

    points: List[Tuple[float, float]]

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYLINE


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox3D(Annotation):
    """
    Three-dimensional cuboid bounding box in a right-handed coordinate system.

    Specified by (x, y, z) coordinates for the ``center`` of the cuboid, (x, y, z) ``dimensions``, and a ``rotation``
    parameter specifying the degrees of rotation about each axis (x, y, z) ranging [-pi, pi].
    """

    #: (x, y, z) coordinates specifying the center of the bounding box.
    center: Tuple[float, float, float]

    #: (x, y, z) measurements specifying the dimensions of the bounding box.
    dimensions: Tuple[float, float, float]

    #: Rotations in degrees about each (x, y, z) axis.
    rotations: Tuple[float, float, float]

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX_3D


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox3D(BoundingBox3D):
    """:class:`BoundingBox3D` with an additional string label."""

    label: str


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox3D(BoundingBox3D):
    """:class:`BoundingBox3D` with an additional float score."""

    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox3D(BoundingBox3D):
    """:class:`BoundingBox3D` with an additional string label and float score."""

    label: str
    score: float


@dataclass(frozen=True, config=ValidatorConfig)
class SegmentationMask(Annotation):
    """
    Raster segmentation mask. The ``locator`` is the URL to the image file representing the segmentation mask.

    The segmentation mask must be rendered as a single-channel, 8-bit-depth (grayscale) image. For the best results,
    use a lossless file format such as PNG. Each pixel's value is the numerical ID of its class label, as specified in
    the ``labels`` map. Any pixel value not present in the ``labels`` map is rendered as part of the background.

    For example, ``labels = {255: "object"}`` will highlight all pixels with the value of 255 as ``"object"``. Every
    other pixel value will be transparent.
    """

    #: Mapping of unique label IDs (pixel values) to unique label values.
    labels: Dict[int, str]

    #: URL of the image (segmentation mask).
    locator: str

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.SEGMENTATION_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class BitmapMask(Annotation):
    """Arbitrary bitmap mask. The ``locator`` is the URL to the image file representing the mask."""

    #: URL of the bitmap data.
    locator: str

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BITMAP_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class ClassificationLabel(Annotation):
    """Label of classification."""

    label: str

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.CLASSIFICATON_LABEL


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredClassificationLabel(ClassificationLabel):
    """Classification label with accompanying score."""

    score: float


_ANNOTATION_TYPES = [
    BoundingBox,
    LabeledBoundingBox,
    ScoredBoundingBox,
    ScoredLabeledBoundingBox,
    Polygon,
    LabeledPolygon,
    ScoredPolygon,
    ScoredLabeledPolygon,
    Keypoints,
    Polyline,
    BoundingBox3D,
    LabeledBoundingBox3D,
    ScoredBoundingBox3D,
    ScoredLabeledBoundingBox3D,
    SegmentationMask,
    BitmapMask,
    ClassificationLabel,
    ScoredClassificationLabel,
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noreorder
from .inference import Inference
from .inference import InferenceType
from .test_image import BaseTestImage
from .test_case import BaseTestCase
from .test_suite import BaseTestSuite
from .model import BaseModel
from .test_run import BaseTestRun
from .test_config import TestConfig

__all__ = [
    "InferenceType",
    "BaseTestCase",
    "BaseTestSuite",
    "BaseTestImage",
    "BaseTestRun",
    "Inference",
    "BaseModel",
    "TestConfig",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Type

from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow import Composite
from kolena.workflow import TestSample
from kolena.workflow._datatypes import DataObject
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import safe_issubclass
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_field
from kolena.workflow.test_sample import _get_composite_fields


@dataclass(frozen=True, config=ValidatorConfig)
class GroundTruth(DataObject):
    """
    The ground truth against which a model is evaluated.

    A test case contains one or more :class:`kolena.workflow.TestSample` objects each paired with a ground truth object.
    During evaluation, these test samples, ground truths, and your model's inferences are provided to the
    :class:`kolena.workflow.Evaluator` implementation.

    This object may contain any combination of scalars (e.g. ``str``, ``float``),
    :class:`kolena.workflow.annotation.Annotation` objects, or lists of these objects.

    For :class:`kolena.workflow.Composite`, each object can contain multiple basic test sample elements.
    To associate a set of attributes and/or annotations as the ground truth to a target test sample element,
    one can use :class:`kolena.workflow.DataObject` and use the same name as in :class:`kolena.workflow.Composite`.

    Continue with the example given in :class:`kolena.workflow.Composite`, which takes an image pair as a
    test sample, one can design ground truth as:

    .. code-block:: python

        class FacePairSample(kolena.workflow.Composite):
            source: Image
            target: Image


        class FaceRegion(DataObject):
            bounding_box: BoundingBox
            keypoints: Keypoints


        class FacePair(GroundTruth):
            source: FaceRegion
            target: FaceRegion
            is_same_person: bool

    making it clear which bounding boxes and keypoints are associated to which image in the test sample.
    """


def _validate_ground_truth_type(
    test_sample_type: Type[TestSample],
    ground_truth_type: Type[GroundTruth],
) -> None:
    if not issubclass(ground_truth_type, GroundTruth):
        raise ValueError(f"Ground truth must subclass {GroundTruth.__name__}")

    is_composite = issubclass(test_sample_type, Composite)
    composite_fields = _get_composite_fields(test_sample_type) if is_composite else []

    for field_name, field_value in get_data_object_field_types(ground_truth_type).items():
        if field_name in composite_fields and safe_issubclass(field_value, DataObject):
            validate_data_object_type(field_value)
        else:
            validate_field(field_name, field_value)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noreorder
from ._datatypes import DataObject
from .test_sample import Metadata
from .test_sample import Image
from .test_sample import ImagePair
from .test_sample import ImageText
from .test_sample import TestSample
from .test_sample import Composite
from .test_sample import Text
from .test_sample import BaseVideo
from .test_sample import Video
from .test_sample import Document
from .ground_truth import GroundTruth
from .inference import Inference
from .workflow import Workflow
from .test_case import TestCase
from .test_suite import TestSuite
from .model import Model
from .evaluator import AxisConfig
from .evaluator import Plot
from .evaluator import Curve
from .evaluator import CurvePlot
from .evaluator import ConfusionMatrix
from .evaluator import Histogram
from .evaluator import BarPlot
from .evaluator import MetricsTestCase
from .evaluator import MetricsTestSample
from .evaluator import MetricsTestSuite
from .evaluator import Evaluator
from .evaluator import EvaluatorConfiguration
from .evaluator_function import BasicEvaluatorFunction as _BasicEvaluatorFunction
from .evaluator_function import TestCases
from .evaluator_function import EvaluationResults
from .test_run import TestRun
from .test_run import test
from ._helpers import define_workflow

__all__ = [
    "DataObject",
    "Metadata",
    "Image",
    "ImagePair",
    "ImageText",
    "TestSample",
    "Composite",
    "Text",
    "BaseVideo",
    "Video",
    "Document",
    "GroundTruth",
    "Inference",
    "Workflow",
    "TestCase",
    "TestSuite",
    "Model",
    "AxisConfig",
    "Plot",
    "Curve",
    "CurvePlot",
    "ConfusionMatrix",
    "Histogram",
    "BarPlot",
    "MetricsTestCase",
    "MetricsTestSample",
    "MetricsTestSuite",
    "Evaluator",
    "EvaluatorConfiguration",
    "BasicEvaluatorFunction",
    "TestCases",
    "EvaluationResults",
    "TestRun",
    "test",
    "define_workflow",
]

# """Copied directly from evaluator_function.py for documentation purposes."""
#: ``kolena.workflow.BasicEvaluatorFunction`` introduces a function based evaluator implementation that takes
#: the inferences for all test samples in a test suite and a :class:`kolena.workflow.TestCases` as input, and computes
#: the corresponding test-sample-level, test-case-level, and test-suite-level metrics (and optionally plots) as output.
#:
#: The control flow is in general more streamlined than with :class:`kolena.workflow.Evaluator`, but requires a couple
#: of assumptions to hold:
#:
#: - Test-sample-level metrics do not vary by test case
#: - Ground truths corresponding to a given test sample do not vary by test case
#:
#: This ``BasicEvaluatorFunction`` is provided to the test run at runtime, and is expected to have the
#: following signature:
#:
#: :param List[kolena.workflow.TestSample] test_samples: A list of distinct :class:`kolena.workflow.TestSample` values
#:     that correspond to all test samples in the test run.
#: :param List[kolena.workflow.GroundTruth] ground_truths: A list of :class:`kolena.workflow.GroundTruth` values
#:     corresponding to and sequenced in the same order as ``test_samples``.
#: :param List[kolena.workflow.Inference] inferences: A list of :class:`kolena.workflow.Inference` values corresponding
#:     to and sequenced in the same order as ``test_samples``.
#: :param TestCases test_cases: An instance of :class:`kolena.workflow.TestCases`, generally used to provide iteration
#:        groupings for evaluating test-case-level metrics.
#: :param EvaluatorConfiguration evaluator_configuration: The configuration to use when performing the evaluation.
#:     This parameter may be omitted in the function definition if running with no configuration.
#: :rtype: :class:`kolena.workflow.EvaluationResults`
#: :return: An object tracking the test-sample-level, test-case-level and test-suite-level metrics and plots for the
#:     input collection of test samples.
BasicEvaluatorFunction = _BasicEvaluatorFunction



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kolena',
 'kolena._api',
 'kolena._api.v1',
 'kolena._utils',
 'kolena._utils.dataframes',
 'kolena.classification',
 'kolena.classification.multiclass',
 'kolena.detection',
 'kolena.detection._internal',
 'kolena.fr',
 'kolena.workflow',
 'kolena.workflow.metrics']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.1,<10.0.0',
 'Shapely>=1.8.5,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'dacite>=1.6',
 'deprecation>=2.1.0,<3.0.0',
 'pandera>=0.9.0',
 'pyarrow>=8',
 'pydantic>=1.8',
 'requests-toolbelt',
 'requests>=2.20,<2.30',
 'retrying>=1.3.3,<2.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'tqdm>=4,<5']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata<5.0',
                             'typing-extensions>=4.5.0,<5.0.0'],
 ':python_version >= "3.11"': ['numpy>=1.23', 'pandas>=1.5,<1.6'],
 ':python_version >= "3.7" and python_version < "3.11"': ['numpy>=1.19',
                                                          'pandas>=1.1,<1.6']}

entry_points = \
{'console_scripts': ['kolena = kolena._utils.cli:run']}

setup_kwargs = {
    'name': 'kolena',
    'version': '0.72.0',
    'description': "Client for Kolena's machine learning (ML) testing and debugging platform.",
    'long_description': '<p align="center">\n  <img src="https://app.kolena.io/api/developer/docs/html/_static/wordmark-purple.svg" width="400" alt="Kolena" />\n</p>\n\n<p align=\'center\'>\n  <a href="https://pypi.python.org/pypi/kolena"><img src="https://img.shields.io/pypi/v/kolena" /></a>\n  <a href="https://www.apache.org/licenses/LICENSE-2.0"><img src="https://img.shields.io/pypi/l/kolena" /></a>\n  <a href="https://github.com/kolenaIO/kolena/actions"><img src="https://img.shields.io/github/checks-status/kolenaIO/kolena/trunk" /></a>\n  <a href="https://codecov.io/gh/kolenaIO/kolena" ><img src="https://codecov.io/gh/kolenaIO/kolena/branch/trunk/graph/badge.svg?token=8WOY5I8SF1"/></a>\n  <a href="https://docs.kolena.io"><img src="https://img.shields.io/badge/resource-docs-6434c1" /></a>\n</p>\n\n---\n\n[Kolena](https://www.kolena.io) is a comprehensive machine learning testing and debugging platform to surface hidden\nmodel behaviors and take the mystery out of model development. Kolena helps you:\n\n- Perform high-resolution model evaluation\n- Understand and track behavioral improvements and regressions\n- Meaningfully communicate model capabilities\n- Automate model testing and deployment workflows\n\nThis `kolena` package contains the Python client library for programmatic interaction with the Kolena ML testing\nplatform.\n\n## Documentation\n\nVisit [docs.kolena.io](https://docs.kolena.io/) for tutorial and usage documentation and the\n[API Reference](https://app.kolena.io/api/developer/docs/html/index.html) for detailed `kolena` typing and\nfunction documentation.\n',
    'author': 'Kolena Engineering',
    'author_email': 'eng@kolena.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://kolena.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.12',
}


setup(**setup_kwargs)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Type

from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow import Composite
from kolena.workflow import TestSample
from kolena.workflow._datatypes import DataObject
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import safe_issubclass
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_field
from kolena.workflow.test_sample import _get_composite_fields


@dataclass(frozen=True, config=ValidatorConfig)
class Inference(DataObject):
    """
    The inference produced by a model.

    Typically the structure of this object closely mirrors the structure of the :class:`kolena.workflow.GroundTruth` for
    a workflow, but this is not a requirement.

    During evaluation, the :class:`kolena.workflow.TestSample` objects, ground truth objects, and these inference
    objects are provided to the :class:`kolena.workflow.Evaluator` implementation to compute metrics.

    This object may contain any combination of scalars (e.g. ``str``, ``float``),
    :class:`kolena.workflow.annotation.Annotation` objects, or lists of these objects.

    A model processing a :class:`kolean.workflow.Composite` object can produce an inference result for each of its
    element. To associate an inference result to each test sample element, one can put the attributes and/or annotations
    inside an :class:`kolena.workflow.DataObject` and use the same name as that in :class:`kolena.workflow.Composite`.

    Continue with the example given in :class:`kolena.workflow.Composite`, which takes an image pair as a
    test sample, one can design inference as:

    .. code-block:: python

        class FacePairSample(kolena.workflow.Composite):
            source: Image
            target: Image


        class FaceRegion(DataObject):
            bounding_box: BoundingBox
            keypoints: Keypoints


        class FacePair(Inference):
            source: FaceRegion
            target: FaceRegion
            similarity: float

    This way, it is clear which bounding boxes and keypoints are associated to which image in the test sample.
    """


def _validate_inference_type(test_sample_type: Type[TestSample], inference_type: Type[Inference]) -> None:
    if not issubclass(inference_type, Inference):
        raise ValueError(f"Inference must subclass {Inference.__name__}")

    is_composite = issubclass(test_sample_type, Composite)
    composite_fields = _get_composite_fields(test_sample_type) if is_composite else []

    for field_name, field_value in get_data_object_field_types(inference_type).items():
        if field_name in composite_fields and safe_issubclass(field_value, DataObject):
            validate_data_object_type(field_value)
        else:
            validate_field(field_name, field_value)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import conlist
from pydantic import constr
from pydantic import StrictBool
from pydantic import StrictStr
from pydantic.dataclasses import dataclass
from typing_extensions import Literal  # for python <3.8 compatibility

from kolena._api.v1.batched_load import BatchedLoad


class Model:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        metadata: Dict[str, Any]
        workflow: str

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        metadata: Dict[str, Any]
        workflow: str

    @dataclass(frozen=True)
    class DeleteRequest:
        id: int


@dataclass(frozen=True)
class TestCaseInfo:
    id: int
    name: str
    sample_count: int
    # provided by stratification preview to show what bucket this test-case falls into,
    # e.g. {"count": "low", "gender": "female"}. This info is not available from stratification result.
    membership: Optional[Dict[str, str]] = None


class TestCase:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str

    @dataclass(frozen=True)
    class CreateFromExistingRequest(BatchedLoad.WithLoadUUID):
        test_case_name: str
        test_suite_name: str
        workflow: str
        compute_metrics_where_possible: bool = False
        compute_metrics_models: Optional[List[int]] = None

    @dataclass(frozen=True)
    class CreateFromExistingResponse:
        test_case_id: int
        test_suite_id: int  # newly created version ID

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        workflow: str

    @dataclass(frozen=True)
    class LoadContentsRequest:
        test_case_id: int

    @dataclass(frozen=True)
    class InitLoadContentsRequest(LoadContentsRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class EditRequest:
        test_case_id: int
        current_version: int
        description: str
        reset: bool = False

    @dataclass(frozen=True)
    class CompleteEditRequest(EditRequest, BatchedLoad.WithLoadUUID):
        ...

    @dataclass(frozen=True)
    class BulkCreateFromExistingRequest:
        task_handle: str
        workflow: str
        test_suite_name: str
        test_suite_description: str
        test_case_names: List[str]
        uuids: List[str]
        compute_metrics_where_possible: bool = False
        compute_metrics_models: Optional[List[int]] = None

    @dataclass(frozen=True)
    class BulkCreateFromExistingResult:
        test_suite_id: int  # ID of specific version
        test_suite_name: str
        test_suite_description: Optional[str]
        test_cases: List[TestCaseInfo]


class TestSuite:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class LoadAllRequest:
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        test_cases: List[TestCase.EntityData]
        workflow: str
        tags: List[str]

    @dataclass(frozen=True)
    class LoadAllResponse:
        test_suites: List["TestSuite.EntityData"]

    @dataclass(frozen=True)
    class EditRequest:
        test_suite_id: int  # ID of version being edited
        current_version: int
        name: str
        description: str
        test_case_ids: List[int]
        tags: Optional[List[str]] = None  # unique set -- list used to preserve ordering

    @dataclass(frozen=True)
    class DeleteRequest:
        test_suite_id: int


TestSuite.LoadAllResponse.__pydantic_model__.update_forward_refs()


class TestRun:
    @dataclass(frozen=True)
    class MarkCrashedRequest:
        test_run_id: int


CATEGORICAL = "categorical"
NUMERIC = "numeric"


@dataclass(frozen=True)
class Dimension:
    column: Literal["test_sample", "test_sample_metadata", "ground_truth"]
    field: str
    datatype: Literal["categorical", "numeric"]

    def is_categorical(self) -> bool:
        return self.datatype == CATEGORICAL

    def is_numeric(self) -> bool:
        return self.datatype == NUMERIC


@dataclass(frozen=True)
class CategoricalBucket:
    values: Union[List[StrictStr], List[StrictBool]]


class IntervalType(str, Enum):
    RIGHT_OPEN = "right-open"
    CLOSED = "closed"


@dataclass(frozen=True)
class NumericBucket:
    min: float
    max: float
    interval_type: IntervalType = IntervalType.RIGHT_OPEN


@dataclass(frozen=True)
class DimensionSpec:
    # arbitrarily selected max_length for sanity reason
    name: constr(min_length=1, max_length=100)
    columns: conlist(Dimension, min_items=1, max_items=3)
    buckets: Dict[str, List[Union[CategoricalBucket, NumericBucket]]]

    def __post_init_post_parse__(self) -> None:
        n_cols = len(self.columns)
        # auto-stratification only supported for single categorical field
        if not self.buckets and not self.should_auto_stratify():
            raise ValueError("Empty bucket.")

        check_duplicate_fields(self.columns)

        for bucket_name, filters in self.buckets.items():
            if len(filters) != n_cols:
                raise ValueError(f"Bucket '{bucket_name}' does not match number of columns {n_cols}.")

            for dimension, filter in zip(self.columns, filters):
                if (dimension.is_categorical() and not isinstance(filter, CategoricalBucket)) or (
                    dimension.is_numeric() and not isinstance(filter, NumericBucket)
                ):
                    raise ValueError(f"Column {dimension} has incompatible filter.")

    def should_auto_stratify(self) -> bool:
        return not self.buckets and len(self.columns) == 1 and self.columns[0].is_categorical()


@dataclass(frozen=True)
class BaseStratifyRequest:
    workflow: str
    test_case_id: int
    test_suite_name: constr(strict=True, min_length=1)
    # stratification should be non-empty, and currently restrict to max 3 cross-dimension
    strata: conlist(DimensionSpec, min_items=1, max_items=3)

    def __post_init_post_parse__(self) -> None:
        unique_names = {s.name for s in self.strata}
        if len(unique_names) != len(self.strata):
            raise ValueError("Duplicate stratification name")

        check_duplicate_fields([stratum for s in self.strata for stratum in s.columns])


@dataclass(frozen=True)
class StratifyResponse:
    test_suite_id: int  # newly created version ID
    test_suite_name: str
    test_suite_description: str
    base_test_case: TestCaseInfo
    stratified_test_cases: List[TestCaseInfo]


def check_duplicate_fields(strata: List[Dimension]) -> None:
    field_set = set()
    for stratum in strata:
        field = (stratum.column, stratum.field)
        # prevent self-cross stratification, i.e. one field can only be used once
        if field in field_set:
            raise ValueError(f"{field} is already used for stratification.")
        field_set.add(field)
